(provide 'esop-multi-basic)
;;; Time-stamp: <2013-06-26 18:31:06 hirayama>
;;; 
;;; This program computes an exact minimum ESOP of a given
;;; multiple-output function.  The programming language is Common
;;; Lisp.
;;; See "00readme-esop-multi-basic.txt" for use. 

;;; Comments in this program are written in UTF8 Japanese.  If your
;;; Lisp interpreter signals errors by the Japanese comments, please
;;; remove all the comments. It can be done with the UNIX command: 
;;; "sed -e 's/;.*$//' < esop-multi-basic.lsp > no-comments.lsp", for example. 


;;; f: 論理関数の整数表現。長さ m * 2^n の bit列。
;;; nは入力変数の個数
;;; mは出力数

;;; 使い方:
;;;
;;; 準備
;;; 多出力関数のnとmを設定する。(set-size n m)
;;; 例: (set-size 7 1)
;;; 最小ESOPの表を作る。(set-table n)
;;; 例: (set-table 4)
;;;
;;; 最小ESOPを求める。(get-esop f n)
;;; 例(fが10進数): (get-esop 1615781404510044160 6)
;;; 例(fが16進数): (get-esop #x166c68a06a80c000 6)
;;;
;;; 実験用関数
;;; (esop-from-fxg f0g f1g f2g g n)
;;; 部分関数f0g, f1g, f2g, gからfのESOPを計算する。nはfxgの変数の個数。
;;; 多値として、f0g, f1g, f2g, gのESOPも返す。
;;; 例: (esop-from-fxg 1615781404510044160 7607307959053854848 1408710670243711 9222245999475408895 6)
;;; (esop-to-latex esop)
;;; esop表現をLaTeX表現に変換した式を返す。
;;; *multi*の値を見て、単一出力か多出力かの判断をする。
;;; "\\"が気になるときには、"format t"と組み合わせて表示する。
;;; 例: (format t "~a" (esop-to-latex (get-esop 1615781404510044160 6)))
;;; (esop-to-bint esop)
;;; esopからbintに戻す。
;;; 例: (esop-to-bint (get-esop 1615781404510044160 6))


;;; 大域変数
;;; 
;;; *multi*
;;;     出力数。
;;; *max-num-of-variables*:
;;;     対象とする関数の変数の個数の最大値。
;;; *bit-length-of-func-array*:
;;;     i番目の要素が m*2^i である整数配列(0 <= i <= *max-num-of-variables*)
;;; 	すなわち、i変数関数fの(整数表現の)ビット長を与える。
;;; *num-of-funcs-array*:
;;;     i変数関数の個数(2^(m*2^i))をi番目の要素としてもつ整数配列
;;;     (0 <= i <= *max-num-of-variables*)


(defvar *max-num-of-variables*)
(defvar *multi* 1)
(defvar *bit-length-of-func-array*)
(defvar *num-of-funcs-array*)
(defvar *set-size*)
(defvar *dummy-big-int*)
(defvar *1term-multi-array*)

(defvar *top-var* 9)
(defvar *counter*)

; リテラルの表現形式
(defvar nulllit '()) ; 0, 定数リテラル
(defvar 0lit '(0))   ; \bar{x}, 否定リテラル
(defvar 1lit '(1))   ; x, 肯定リテラル
(defvar clit '(0 1)) ; 1, 定数リテラル
; 注: '(1 0)はclitとみなされない。
(defvar *zero-cvlis*
  ;; 定数0関数のcvlis表現
  nulllit)

(defun expt2 (i) (ash 1 i)) ; 2^i を求める。

(defun set-size (n m &key (order 'down))
  ;;よく使う数値をあらかじめ配列に設定しておく。
  (setq *max-num-of-variables* n)
  (setq *multi* m)
  (setq *bit-length-of-func-array*
	(make-array (1+ *max-num-of-variables*)
		    :initial-contents
		    (do ((i *max-num-of-variables* (1- i))
			 (rtn () (cons (* m (expt2 i)) rtn)))
			((< i 0) rtn))))
  (setq *num-of-funcs-array*
	(make-array (1+ *max-num-of-variables*)
		    :initial-contents
		    (do ((i *max-num-of-variables* (1- i))
			 (rtn ()
			      (cons (expt2 (* m (expt2 i)))
				    rtn)))
			((< i 0) rtn))))
  (setq *set-size* n)
  (setq *dummy-big-int* (1+ (svref *bit-length-of-func-array* n)))
  (set-1term-multi (1- n) m :order order)
  ;; 最小ESOPの表を準備
  (labels ((table-n (m)
	     (cond ((= m 1) (min (1- n) 4)) ((<= m 2) (min (1- n) 3)) ((<= m 5) (min (1- n) 2)) ((<= m 11) (min (1- n) 1)) (t 0)))
	   ;; 大容量メモリが使用可能な場合、下記を試してもよい。
	   ;;(cond ((= m 1) (min (1- n) 4)) ((<= m 3) (min (1- n) 3)) ((<= m 6) (min (1- n) 2)) ((<= m 12) (min (1- n) 1)) (t 0)))
	   )
    (format t "Making the table of minimum ESOPs with n=~A. Wait a moment...~%" (table-n m))
    (set-table (table-n m)))
  'ok
  )

;;; operations of ESOP

;;; esop: 多出力関数の特性関数のESOP
;;; 項数、リテラル数、論理式(cvlis)からなるリスト。

(defmacro esop-num-of-terms (esop) `(car ,esop))
(defmacro esop-num-of-literals (esop) `(cadr ,esop))
(defmacro esop-cvlis (esop) `(caddr ,esop))
(defmacro esop-make (n l cvlis) `(list ,n ,l ,cvlis))

;;; cvlis: ESOP(論理式)を表すデータタイプ。
;;; 積項のリストでESOPを表現する。
;;; 積項はリテラルのリストで表現する。

(defun cv/3-to-cvlis (cvlis0 cvlis1 cvlis2)
  ;; (n-1)変数のcvlisからn変数のcvlisを作る。
  (append
   (mapcar #'(lambda (x)
               (cons 0lit x))
           cvlis0)
   (mapcar #'(lambda (x)
               (cons 1lit x))
           cvlis1)
   (mapcar #'(lambda (x)
               (cons clit x))
           cvlis2)
   ))

;(defun cvlis-output-phase (cvlis)
;  ;; 多出力回路の出力位相を求める。
;  (let ((count 0))
;    (dolist (term cvlis count)
;      (incf count (length (car (last term)))))))

;(defun cvlis-num-of-nega-literals (cvlis)
;  ;; 論理式の否定リテラル数を求める。
;  ;; (出力部リテラルはカウントしない)
;  (let ((count 0))
;    (dolist (term cvlis count)
;      (dolist (literal (butlast term) count)
;	(when (equal literal 0lit)
;	  (incf count))))))

;;;
;;; *esop-table*
;;;

(defvar *esop-table-size-base* -1)
(defvar *esop-table* nil)

(defun set-table (n)
  ;; n変数関数の最小ESOPの表を準備する。
  ;; 出力数は*multi*を仮定している。
  (setq *esop-table-size-base* -1) ; 初期化
  (setq *esop-table* nil) ; 初期化
  (setq *esop-table* (make-esop-table n))
  (setq *esop-table-size-base* n)
  )

(defun make-esop-table (n)
  ;; n変数関数の最小ESOPの表を作る。
  ;; 出力数は*multi*とする。
  (cond ((<= n 0)
	 (make-array (svref *num-of-funcs-array* 0)
		     :initial-contents 
		     (initial-esop-list *multi*)))
	((<= n *esop-table-size-base*) *esop-table*)
	(t
	 (let* ((*esop-table* (make-esop-table (1- n)))
		(*esop-table-size-base* (1- n))
		(tbl-size (svref *num-of-funcs-array* n))
		(tbl (make-array tbl-size)))
	   (do ((i (1- tbl-size) (1- i)))
	       ((< i 0) tbl)
	     (setf (svref tbl i) (get-esop i n)))))))

(defun initial-esop-list (m)
  ;; m出力の0変数関数のESOPのリストを返す。(m >= 1)
  ;; (出力部リテラルはリテラル数にカウントしない。)
  (let ((lis nil))
    (do ((i (- (expt2 m) 1) (1- i)))
	((= i 0) lis)
      (push (esop-make 1 0 (list (list (initial-multi-literal i)))) lis))
    (push (esop-make 0 0 *zero-cvlis*) lis)
    lis))

(defun initial-multi-literal (bint)
  ;; m出力の0変数関数のbintのリテラル表現を返す。
  (do ((xbint bint (ash xbint -1))
       (count 0 (1+ count))
       (lis nil))
      ((= xbint 0) (reverse lis))
    (when (= (logand xbint 1) 1)
      (push count lis))))
      
;;;
;;; get-esop
;;;

(defun subfuncs (f n)
  ;; f is an n-varible function
  ;; f の3つの部分関数を返す
  (let* ((f1 (ash f
		  (- (svref *bit-length-of-func-array* (1- n)))))
	 (f0 (logxor f
		     (ash f1 
			  (svref *bit-length-of-func-array* (1- n)))))
	 (f2 (logxor f1 f0)))
    (values f0 f1 f2)))

;(defun esop-< (esop1 esop2)
;  ;; 二つのESOPの項数の大小比較をする。
;  ;; 項数が等しいときは、リテラル数で比較する。
;  (or (< (esop-num-of-terms esop1)
;	 (esop-num-of-terms esop2))
;      (and (= (esop-num-of-terms esop1)
;	      (esop-num-of-terms esop2))
;	   (< (esop-num-of-literals esop1)
;	      (esop-num-of-literals esop2)))))
;(defun esop-<= (esop1 esop2)
;  ;; 二つのESOPの項数の大小比較をする。
;  ;; 項数が等しいときは、リテラル数で比較する。
;  (or (< (esop-num-of-terms esop1)
;	 (esop-num-of-terms esop2))
;      (and (= (esop-num-of-terms esop1)
;	      (esop-num-of-terms esop2))
;	   (<= (esop-num-of-literals esop1)
;	       (esop-num-of-literals esop2)))))
;(defun esop-<= (esop1 esop2)
;  ;; 二つのESOPの項数の大小比較をする。
;  ;; 決着がつかないときの比較内容の優先順位は以下の通り。
;  ;; 第一に項数を比較する。
;  ;; 第二にリテラル数を比較する。
;  ;; 第三に出力位相を比較する。
;  ;; 第四に否定リテラル数を比較する。
;  ;; 上記すべてにおいて等しいときには、Trueを返す。
;  (cond ((< (esop-num-of-terms esop1)
;	    (esop-num-of-terms esop2))
;	 t)
;	((> (esop-num-of-terms esop1)
;	    (esop-num-of-terms esop2))
;	 nil)
;	((< (esop-num-of-literals esop1)
;	    (esop-num-of-literals esop2))
;	 t)
;	((> (esop-num-of-literals esop1)
;	    (esop-num-of-literals esop2))
;	 nil)
;	((< (cvlis-output-phase (esop-cvlis esop1))
;	    (cvlis-output-phase (esop-cvlis esop2)))
;	 t)
;	((> (cvlis-output-phase (esop-cvlis esop1))
;	    (cvlis-output-phase (esop-cvlis esop2)))
;	 nil)
;	((< (cvlis-num-of-nega-literals (esop-cvlis esop1))
;	    (cvlis-num-of-nega-literals (esop-cvlis esop2)))
;	 t)
;	((> (cvlis-num-of-nega-literals (esop-cvlis esop1))
;	    (cvlis-num-of-nega-literals (esop-cvlis esop2)))
;	 nil)
;	(t
;	 t)))

;(defun select3 (esop0 esop1 esop2 esop-g)
;  ;; 4つESOPのうち、項数の少ない3つを選んで、順序を直して、多値で返す。
;  ;; faster
;  (if (esop-<= esop0 esop1)
;      (if (esop-<= esop1 esop2)
;	   (if (esop-<= esop2 esop-g)
;	       (values esop1 esop0 esop2)
;	       (values esop0 esop1 esop-g))
;	   (if (esop-<= esop1 esop-g)
;	       (values esop1 esop0 esop2)
;	       (values esop-g esop2 esop0)))
;      (if (esop-<= esop0 esop2)
;	   (if (esop-<= esop2 esop-g)
;	       (values esop1 esop0 esop2)
;	       (values esop0 esop1 esop-g))
;	   (if (esop-<= esop0 esop-g)
;	       (values esop1 esop0 esop2)
;	       (values esop2 esop-g esop1)))))

(defun select3 (esop0 esop1 esop2 esop-g)
  ;; 4つESOPのうち、項数の少ない3つを選んで、順序を直して、多値で返す。
  ;; 項数が等しいときには、連結ESOPになったときのリテラル数が少ないものを選ぶ。
  ;; slower than the above
  (let ((min-tau *dummy-big-int*)
	(min-literal *dummy-big-int*)
	(result nil)
	(triple-list (list (list esop1 esop0 esop2)
			   (list esop0 esop1 esop-g)
			   (list esop1 esop0 esop2)
			   (list esop-g esop2 esop0)
			   (list esop1 esop0 esop2)
			   (list esop0 esop1 esop-g)
			   (list esop1 esop0 esop2)
			   (list esop2 esop-g esop1))))
    (dolist (triple triple-list result)
      (let* ((xtau (apply #'nterms-of-subesops triple))
	     (xliteral (apply #'nliterals-of-subesops triple)))
	(if (or (< xtau min-tau)
		(and (= xtau min-tau)
		     (< xliteral min-literal)))
	    (progn
	      (setq min-tau xtau)
	      (setq min-literal xliteral)
	      (setq result triple)))))
    (values-list result)))

(defun nterms-of-subesops (esop0 esop1 esop2)
  ;; 3つの(n-1)変数ESOPを取り、1つのn変数ESOP
  ;; \bar{x}_n (esop0) \oplus x_n (esop1) \oplus (esop2)
  ;; にしたときの積項数を返す。
  (+ (esop-num-of-terms esop0)
     (esop-num-of-terms esop1)
     (esop-num-of-terms esop2)))

(defun nliterals-of-subesops (esop0 esop1 esop2)
  ;; 3つの(n-1)変数ESOPを取り、1つのn変数ESOP
  ;; \bar{x}_n (esop0) \oplus x_n (esop1) \oplus (esop2)
  ;; にしたときのリテラル数を返す。
  (+ (esop-num-of-literals esop0)
     (esop-num-of-terms esop0)
     (esop-num-of-literals esop1)
     (esop-num-of-terms esop1)
     (esop-num-of-literals esop2)))

(defun get-esop-with-g (f0+g+p f1+g+p f2+g+p g+p n esop-g+p)
  ;; T(f,g+p)に相当する手続き
  ;; f0+g+p, f1+g+p, f2+g+p, gはn変数関数。(pは積項関数)
  ;; 出力数は、変数*multi*に設定されていることを想定している。
  ;; 与えられた4つの関数のESOPを求めて、
  ;; 値の小さい３つを順序を直して多値で返す。
  ;; 処理高速化のため、ここではesop-makeしない。
  g+p ; reserved
  (let* ((esop-f0+g+p (get-esop f0+g+p n))
	 (esop-f1+g+p (get-esop f1+g+p n))
	 (esop-f2+g+p (get-esop f2+g+p n)))
    (select3 esop-f0+g+p esop-f1+g+p esop-f2+g+p esop-g+p)))


(defmacro k-upperbound (smin max-fx)
  ;; 探索範囲 k の上界を求める。
  `(min (floor (/ (1- ,smin) 3)) (1- (- ,smin ,max-fx))))

(defun get-esop (f n)
  ;; n変数関数fのesopを求める。
  ;; 出力数は*multi*を仮定している。
  (let ((smin nil)
	(k-upper nil)
	(max-fx nil)
	(n-1 (1- n)))
    (labels ((get-esop-main (f0+g f1+g f2+g g plis tau-fg tau-g)
		;; 論文の手続きSに相当する局所関数。
		;; f0+g, f1+g, f2+gはfの部分関数に候補関数gをEXORした関数。
		;; plisは積項関数の集合、
		;; 単純に深さ優先探索をする。
		(if (or (>= (- tau-fg (- k-upper tau-g))
			    (esop-num-of-terms smin))
			(null plis))
		    smin
		    (let* ((xtau nil)
			   (xnliteral nil)
			   (xcvlis nil)
			   (p (car plis))
			   (rest-plis (cdr plis))
			   (f0+g+p (logxor f0+g p))
			   (f1+g+p (logxor f1+g p))
			   (f2+g+p (logxor f2+g p))
			   (esop0 nil)
			   (esop1 nil)
			   (esop2 nil)
			   (g+p (logxor g p))
			   (esop-g+p (get-esop g+p n-1))
			   (tau-g+p (esop-num-of-terms esop-g+p)))
		      (if (> tau-g+p tau-g) ; (= tau-g+p (1+ tau-g)) の代わり
			  (progn
			    (multiple-value-setq (esop0 esop1 esop2)
			      (get-esop-with-g f0+g+p f1+g+p f2+g+p g+p n-1 esop-g+p))
			    (setq xtau (nterms-of-subesops esop0 esop1 esop2))
			    (setq xnliteral (nliterals-of-subesops esop0 esop1 esop2))
			    (if (or (< xtau (esop-num-of-terms smin))
				    (and (= xtau (esop-num-of-terms smin))
					 (< xnliteral
					    (esop-num-of-literals smin)))) ; 良い解か
				(progn
				  (setq xcvlis (cv/3-to-cvlis (esop-cvlis esop0)
							      (esop-cvlis esop1)
							      (esop-cvlis esop2)))
				  (setq smin (esop-make xtau xnliteral xcvlis))
				  (setq k-upper (k-upperbound (esop-num-of-terms smin) max-fx))))
			    ;; 深さ優先探索
			    (get-esop-main f0+g+p f1+g+p f2+g+p g+p rest-plis xtau tau-g+p)))
		      ;; 幅探索
		      (get-esop-main f0+g f1+g f2+g g rest-plis tau-fg tau-g)))))
	    ;; end of definition of local functions
	    (if (<= n *esop-table-size-base*)
		(svref *esop-table* f)
		(multiple-value-bind (f0 f1 f2) (subfuncs f n)
		  (let* ((term-lis (svref *1term-multi-array* n-1))
			 (g 0)
			 (esop-g (esop-make 0 0 *zero-cvlis*))
			 (esop-f0 (get-esop f0 n-1))
			 (esop-f1 (get-esop f1 n-1))
			 (esop-f2 (get-esop f2 n-1))
			 (esop0 nil)
			 (esop1 nil)
			 (esop2 nil)
			 (xtau nil)
			 (xnliteral nil)
			 (xcvlis nil))
		    (multiple-value-setq (esop0 esop1 esop2)
		      (select3 esop-f0 esop-f1 esop-f2 esop-g))
		    (setq xtau (nterms-of-subesops esop0 esop1 esop2))
		    (setq xnliteral (nliterals-of-subesops esop0 esop1 esop2))
		    (setq xcvlis (cv/3-to-cvlis (esop-cvlis esop0)
						(esop-cvlis esop1)
						(esop-cvlis esop2)))
		    (setq smin (esop-make xtau xnliteral xcvlis))
		    (setq max-fx (max (esop-num-of-terms esop0)
				      (esop-num-of-terms esop1)
				      (esop-num-of-terms esop2)))
		    (setq k-upper (k-upperbound (esop-num-of-terms smin) max-fx))
		    (get-esop-main f0 f1 f2 g term-lis xtau 0)
		    smin))))))
		  
;;;
;;; (多出力関数の)積項リストを生成するための関数
;;;

(defun set-1term-multi (n m &key (order nil))
  ;; あらかじめ積項リストを計算しておいて、配列に格納する。
  (setq *1term-multi-array* (make-array (1+ n)))
  (setf (svref *1term-multi-array* 0) (make-1term-multi 0 m))
  (do ((i 1 (1+ i)))
      ((= i (1+ n)) 'ok)
      (let ((result (mapcan #'(lambda (term) (extend-term term i m))
			    (svref *1term-multi-array* (1- i)))))
	(if order ; 積項リストをソートしたい場合
	    (setf (svref *1term-multi-array* i) 
		  (sort-1term-list result :order order))
	    (setf (svref *1term-multi-array* i) 
		  result)))))

(defun extend-term (term n m)
  ;; (n-1)変数積項termから、３つのn変数積項関数を生成してリストで返す。
  ;; termにつけるx_nのリテラルとしては、
  ;; 否定リテラル、肯定リテラル、定数1リテラルの３とおりである。
  ;; (termは関数0ではないとする)
  m ; (reserved)
  (let* ((bit-length-of-subfunc (* (expt2 (1- n)) m))
	 (neg-term term)
	 (pos-term (ash term bit-length-of-subfunc))
	 (con-term (logxor neg-term pos-term)))
    (list con-term pos-term neg-term))) ; 並び順には意味はない

(defun make-1term-multi (n m)
  ;; n変数積項関数のリストを求める。
  ;; (関数0は積項関数でないとする。)
  (if (= n 0)
      (make-m-literal-list m)
      (mapcan #'(lambda (term)
		  (extend-term term n m))
	      (make-1term-multi (1- n) m))))

(defun make-m-literal-list (m)
  ;; (0,m)関数のうち、関数0以外のすべての関数をリストにして返す。(m >= 1)
  (let ((result nil))
    (dotimes (i (expt2 m) (cdr (reverse result)))
      (push i result))))
	
;;; 1termのリストをsortするときに使う

(defun sort-1term-list (1term-list &key (order 'up))
  ;; 1term-listをセル数順にソートしたリストを返す。
  ;; :orderキーに'upまたは'downを指定することで、
  ;; それぞれ昇順または降順になる。
  ;; セル数が等しい積項があるので、ソートの結果は一意ではない。
  (cond ((equal order 'up)
	 (sort 1term-list #'bint-area-leq))
	((equal order 'down)
	 (sort 1term-list #'bint-area-geq))
	(t
	 (error "`:order' should be 'up or 'down "))))

(defun bint-area-leq (bint1 bint2)
  ;; bint1とbint2のセル数（最小積項数）を比較し、
  ;; bint1 <= bint2 かどうか返す。
  (<= (logcount bint1) (logcount bint2)))

(defun bint-area-geq (bint1 bint2)
  ;; bint1とbint2のセル数（最小積項数）を比較し、
  ;; bint1 >= bint2 かどうか返す。
  (>= (logcount bint1) (logcount bint2)))


;;;
;;; 以下は実験用関数。
;;;

(defun print-debug-data (n tau-g len smin tau-fgp)
  ;; デバッグデータの出力
  (format t "~%~a n=~a~a k=~a, rest=~a, tau-fgp=~a~a, lower-bound=~a."
	  (make-space (* (- n 5) 2))
	  n
	  (make-space tau-g)
	  (1+ tau-g)
	  len
	  tau-fgp
	  (if (= smin tau-fgp) "*" "")
	  (- tau-fgp (- (floor (/ (1- smin) 3)) tau-g))))

(defun esop-to-latex (esop)
  ;; esop表現をLaTeX表現に変換する。
  ;; *multi*の値を見て、単一出力か多出力かの判断をする。
  (let* ((cvlis (esop-cvlis esop))
	 (latex-esop '()))
    (dolist (term cvlis (apply #'concatenate
			       'string
			       (nreverse (cdr latex-esop))))
      (push (term-to-latex term)
	    latex-esop)
      (push " \\oplus"
	    latex-esop))))

(defun term-to-latex (term)
  ;; esop-to-latexの補助関数
  ;; 積項をLaTeX表現に変換する。
  ;; *multi*の値を見て、単一出力か多出力かの判断をする。
  (let* ((literals (butlast term))
	 (out-literal (car (last term)))
	 (latex-term '()))
    (do* ((count (length literals) (1- count))
	  (xliterals literals (cdr xliterals))
	  (literal (car xliterals) (car xliterals)))
	((null xliterals) 'ok)
      (cond ((equal literal 0lit)
	     (push (format nil " \\bar{x}_{~a}" count)
		   latex-term))
	    ((equal literal 1lit)
	     (push (format nil " x_{~a}" count)
		   latex-term))
	    ((equal literal clit)
	     nil)))
    (if (and (null latex-term)
	     (not (null out-literal))) ; constant term
	(push " 1" latex-term))
    (if (>= *multi* 2) ; multiple-output esop
	(push (format nil " y^{\\{~a\\}}" (list-to-string out-literal))
	      latex-term))
    (apply #'concatenate
	   'string
	   (nreverse latex-term))))
    
(defun list-to-string (lis)
  (if (null (cdr lis))
      (format nil "~a" (car lis))
      (concatenate 'string (format nil "~a," (car lis))
		   (list-to-string (cdr lis)))))

(defun esop-from-fxg (f0g f1g f2g g n)
  ;; f0g, f1g, f2g, gからfのESOPを計算する。
  ;; nはfxgの変数の個数。
  ;; 多値として、f0g, f1g, f2g, gのESOPも返す。
  (let ((esop-f0g (get-esop f0g n))
	(esop-f1g (get-esop f1g n))
	(esop-f2g (get-esop f2g n))
	(esop-g (get-esop g n)))
    (multiple-value-bind (esop0 esop1 esop2)
	(select3 esop-f0g esop-f1g esop-f2g esop-g)
      (values 
       (esop/3-to-esop esop0 esop1 esop2)
       esop-f0g esop-f1g esop-f2g esop-g))))

(defun esop/3-to-esop (esop0 esop1 esop2)
  ;; (n-1)変数のesopからn変数のesopを作る。
  (esop-make (nterms-of-subesops esop0 esop1 esop2)
	     (nliterals-of-subesops esop0 esop1 esop2)
	     (cv/3-to-cvlis (esop-cvlis esop0)
			    (esop-cvlis esop1)
			    (esop-cvlis esop2))))

(defun esop-to-bint (esop)
  ;; esopからbintに戻す。
  (let ((result 0))
    (dolist (term (esop-cvlis esop) result)
      (setf result (logxor result (term-to-bint term))))))

(defun term-to-bint (term)
  ;; termからbintに戻す。
  (let ((n (1- (length term))))
    (if (= n 0)
	(outterm-to-bint (car term))
	(let ((subbint (term-to-bint (cdr term))))
	  (expand-literal-termbint (car term) subbint (* (expt2 (1- n)) *multi*))))))
	  
    
(defun outterm-to-bint (outterm)
  ;; outterm部分のみを受取り、bintにする。
  (let ((result 0))
    (dolist (i outterm result)
      (incf result (expt2 i)))))

(defun expand-literal-termbint (literal bint bitlen)
  ;; n変数termbintに(n+1)番目の変数のliteralを拡張したbint値を返す。
  (let ((result 0))
    (dolist (i literal result)
      (setf result (logxor result (ash bint (* i bitlen)))))))


(defun test-all-functions (n)
  ;; get-esopの動作チェック用手続き
  ;; すべての(n,m)関数について、get-esopを実行し、結果をリストで返す。
  (let ((result nil))
    (dotimes (i (svref *num-of-funcs-array* n) (reverse result))
      (push (list i (get-esop i n)) result))))

(defun make-space (n)
  ;; おまけ関数
  ;; 指定した長さの空白を生成する。
  ;; テストモードで使用している。
  (if (<= n 0)
      ""
      (concatenate 'string " " (make-space (1- n)))))

(defun up-down-compare (f n)
  ;; term-listがupのときとdownのときとで解の計算時間が
  ;; どの程度異なるかを確認する。
  (setq *top-var* n)
  (setq *counter* 0)
  (set-1term-multi n *multi* :order 'up)
  (time (get-esop f n))
  (print (list 'up-count *counter*))
  (setq *counter* 0)
  (set-1term-multi n *multi* :order 'down)
  (time (get-esop f n))
  (print (list 'down-count *counter*))
)

