(provide 'tau-multi-basic)
;;; Time-stamp: <2013-06-26 16:12:18 hirayama>
;;; 
;;; This program computes tau (the term count of the exact minimum
;;; ESOP) of a given multiple-output function.  This is the tau
;;; version of esop-multi-basic.lsp. It is much faster than
;;; esop-multi-basic.lsp becasue it computes not the minimum ESOPs
;;; (logic expressions) but the tau only.
;;; The programming language is Common Lisp.
;;; See "00readme-tau-multi-basic.txt" for use.

;;; Comments in this program are written in UTF8 Japanese.  If your
;;; Lisp interpreter signals errors by the Japanese comments, please
;;; remove all the comments. It can be done with the UNIX command
;;; "sed -e 's/;.*$//' < esop-multi-basic.lsp > no-comments.lsp", for example. 


;;; 多出力関数の特性関数のtauを求める。
;;; (このプログラムではESOPではなく、tauのみ求める。)
;;; 使用しているアルゴリズムは、IEICE 2002年12月の論文のmin-tau。
;;; set-size実行後、出力数は変数*multi*で参照される。
;;; (*multi* = 1 にすれば一出力関数の最小化になる)
;;;

;;; f: 論理関数の整数表現。長さ m * 2^n の bit列。
;;; nは入力変数の個数
;;; mは出力数

;;; 使い方:
;;;
;;; 準備
;;; 多出力関数のnとmを設定する。(set-size n m)
;;; 例: (set-size 7 1)
;;; 最小ESOPの表を作る。(set-table n)
;;; 例: (set-table 4)
;;;
;;; tauを求める。(get-tau f n)
;;; 例(fが10進数): (get-tau 1615781404510044160 6)
;;; 例(fが16進数): (get-tau #x166c68a06a80c000 6)

;;; 大域変数
;;; 
;;; *multi*
;;;     出力数。
;;; *max-num-of-variables*:
;;;     対象とする関数の変数の個数の最大値。
;;; *bit-length-of-func-array*:
;;;     i番目の要素が m*2^i である整数配列(0 <= i <= *max-num-of-variables*)
;;; 	すなわち、i変数関数fの(整数表現の)ビット長を与える。
;;; *num-of-funcs-array*:
;;;     i変数関数の個数(2^(m*2^i))をi番目の要素としてもつ整数配列
;;;     (0 <= i <= *max-num-of-variables*)


(defvar *max-num-of-variables*)
(defvar *multi* 1)
(defvar *bit-length-of-func-array*)
(defvar *num-of-funcs-array*)
(defvar *set-size*)
(defvar *dummy-big-int*)
(defvar *1term-multi-array*)

(defvar *top-var*)
(defvar *counter*)

(defun expt2 (i) (ash 1 i)) ; 2^i を求める。

(defun set-size (n m &key (order 'down))
  ;;よく使う数値をあらかじめ配列に設定しておく。
  (setq *max-num-of-variables* n)
  (setq *multi* m)
  (setq *bit-length-of-func-array*
	(make-array (1+ *max-num-of-variables*)
		    :initial-contents
		    (do ((i *max-num-of-variables* (1- i))
			 (rtn () (cons (* m (expt2 i)) rtn)))
			((< i 0) rtn))))
  (setq *num-of-funcs-array*
	(make-array (1+ *max-num-of-variables*)
		    :initial-contents
		    (do ((i *max-num-of-variables* (1- i))
			 (rtn ()
			      (cons (expt2 (* m (expt2 i)))
				    rtn)))
			((< i 0) rtn))))
  (setq *set-size* n)
  (setq *dummy-big-int* (1+ (svref *bit-length-of-func-array* n)))
  (set-1term-multi (1- n) m :order order)
  ;; 最小tauの表を準備
  (labels ((table-n (m)
	     (cond ((= m 1) (min (1- n) 4)) ((<= m 2) (min (1- n) 3)) ((<= m 5) (min (1- n) 2)) ((<= m 12) (min (1- n) 1)) (t 0)))
	   ;; 大容量メモリが使用可能な場合、下記を試してもよい。
	   ;;(cond ((= m 1) (min (1- n) 4)) ((<= m 3) (min (1- n) 3)) ((<= m 6) (min (1- n) 2)) ((<= m 12) (min (1- n) 1)) (t 0)))
	   )
    (format t "Making the tau table of n=~A. Wait a moment...~%" (table-n m))
    (set-table (table-n m)))
  'ok
  )

;;;
;;; *tau-table*
;;;

(defvar *tau-table-size-base* -1)
(defvar *tau-table* nil)

(defun set-table (n)
  ;; n変数関数のtauの表を準備する。
  ;; 出力数は*multi*を仮定している。
  (setq *tau-table-size-base* -1) ; 初期化
  (setq *tau-table* nil) ; 初期化
  (setq *tau-table* (make-tau-table n))
  (setq *tau-table-size-base* n)
  )

(defun make-tau-table (n)
  ;; n変数関数のtauの表を作る。
  ;; 出力数は*multi*とする。
  (cond ((<= n 0)
	 (let ((init-array (make-array (svref *num-of-funcs-array* 0)
				       :initial-element 1)))
	   (setf (svref init-array 0) 0) ; 関数0のみ積項数0である。
	   init-array))
	((<= n *tau-table-size-base*) *tau-table*)
	(t
	 (let* ((*tau-table* (make-tau-table (1- n)))
		(*tau-table-size-base* (1- n))
		(tbl-size (svref *num-of-funcs-array* n))
		(tbl (make-array tbl-size)))
	   (do ((i (1- tbl-size) (1- i)))
	       ((< i 0) tbl)
	     (setf (svref tbl i) (get-tau i n)))))))

;;;
;;; get-tau
;;;

(defun subfuncs (f n)
  ;; f is an n-varible function
  ;; f の3つの部分関数を返す
  (let* ((f1 (ash f
		  (- (svref *bit-length-of-func-array* (1- n)))))
	 (f0 (logxor f
		     (ash f1 
			  (svref *bit-length-of-func-array* (1- n)))))
	 (f2 (logxor f1 f0)))
    (values f0 f1 f2)))

(defun sum-smaller3 (t0 t1 t2 t3)
  ;; 4つのうち、値の小さい3つの和を返す。
  (- (+ t0 t1 t2 t3)
     (max t0 t1 t2 t3)))

(defun get-tau-with-g (f0+g+p f1+g+p f2+g+p g+p n tau-g+p)
  ;; T(f,g+p)に相当する手続き
  ;; f0+g+p, f1+g+p, f2+g+p, gはn変数関数。(pは積項関数)
  ;; tau-g+pはtau(g+p)。
  ;; 出力数は、変数*multi*に設定されていることを想定している。
  ;; tau(f0+g+p), tau(f1+g+p), tau(f2+g+p), tau(g+p)のうち、
  ;; 値の小さい３つを足した和を返す。
  g+p ; reserved
  (let* ((tau-f0+g+p (get-tau f0+g+p n))
	 (tau-f1+g+p (get-tau f1+g+p n))
	 (tau-f2+g+p (get-tau f2+g+p n)))
    (sum-smaller3 tau-f0+g+p tau-f1+g+p tau-f2+g+p tau-g+p)))

(defmacro k-upperbound (smin max-fx)
  ;; 探索範囲 k の上界を求める。
  `(min (floor (/ (1- ,smin) 3)) (1- (- ,smin ,max-fx))))

(defun get-tau (f n)
  ;; n変数関数fのtauを求める。
  ;; 出力数は*multi*を仮定している。
  (let ((smin nil)
	(k-upper nil)
	(max-fx nil)
	(n-1 (1- n)))
    (labels ((get-tau-main (f0+g f1+g f2+g g plis tau-fg tau-g)
		;; 論文の手続きSに相当する局所関数。
		;; f0+g, f1+g, f2+gはfの部分関数に候補関数gをEXORした関数。
		;; plisは積項関数の集合、
		;; tau-fgはT(f,g)。tau-gはtau(g)
		;; 単純に深さ優先探索をする。
		(if (or (>= (- tau-fg (- k-upper tau-g)) smin)
			(null plis))
		    smin
		    (let* ((p (car plis))
			   (rest-plis (cdr plis))
			   (f0+g+p (logxor f0+g p))
			   (f1+g+p (logxor f1+g p))
			   (f2+g+p (logxor f2+g p))
			   (g+p (logxor g p))
			   (tau-fgp nil)
			   (tau-g+p (get-tau g+p n-1)))
		      (if (> tau-g+p tau-g) ; (= tau-g+p (1+ tau-g)) の代わり
			  (progn
			    (setq tau-fgp (get-tau-with-g f0+g+p f1+g+p f2+g+p g+p n-1 tau-g+p))
			    (if (< tau-fgp smin) ; 良い解か
				(progn 
				  (setq smin tau-fgp)
				  (setq k-upper (k-upperbound smin max-fx))))
			    ;; 深さ優先探索
			    (get-tau-main f0+g+p f1+g+p f2+g+p g+p rest-plis tau-fgp tau-g+p)))
		      ;; 幅探索
		      (get-tau-main f0+g f1+g f2+g g rest-plis tau-fg tau-g)))))
	    ;; end of definition of local functions
	    (if (<= n *tau-table-size-base*)
		(svref *tau-table* f)
		(multiple-value-bind (f0 f1 f2) (subfuncs f n)
		  (let* ((term-lis (svref *1term-multi-array* n-1))
			 (g 0)
			 (tau-g 0)
			 (tau-f0 (get-tau f0 n-1))
			 (tau-f1 (get-tau f1 n-1))
			 (tau-f2 (get-tau f2 n-1))
			 (tau-fg (sum-smaller3 tau-f0 tau-f1 tau-f2 tau-g)))
		    (setq smin tau-fg)
		    (setq max-fx (max tau-f0 tau-f1 tau-f2))
		    (setq k-upper (k-upperbound smin max-fx))
		    (get-tau-main f0 f1 f2 g term-lis tau-fg 0)
		    smin))))))

;(defun get-tau (f n)
;  ;; n変数関数fのtauを求める。
;  ;; 出力数は*multi*を仮定している。
;  (let ((smin nil)
;	(k-upper nil)
;	(max-fx nil)
;	(n-1 (1- n)))
;    (labels ((get-tau-main (f0+g f1+g f2+g g plis tau-fg tau-g)
;		;; 論文の手続きSに相当する局所関数。
;		;; f0+g, f1+g, f2+gはfの部分関数に候補関数gをEXORした関数。
;		;; plisは積項関数の集合、
;		;; tau-fgはT(f,g)。tau-gはtau(g)
;		;; 単純に深さ優先探索をする。(繰り返し版)
;		(do* ((rest-plis plis))
;		    ((or (>= (- tau-fg (- k-upper tau-g)) smin)
;			 (null rest-plis))
;		     smin)
;		  (let* ((p (car rest-plis))
;			 (f0+g+p (logxor f0+g p))
;			 (f1+g+p (logxor f1+g p))
;			 (f2+g+p (logxor f2+g p))
;			 (g+p (logxor g p))
;			 (tau-fgp nil)
;			 (tau-g+p (get-tau g+p n-1)))
;		    (setq rest-plis (cdr rest-plis))
;		    (if (> tau-g+p tau-g) ; (= tau-g+p (1+ tau-g)) の代わり
;			(progn
;			  (setq tau-fgp (get-tau-with-g f0+g+p f1+g+p f2+g+p g+p n-1 tau-g+p))
;			  (if (< tau-fgp smin) ; 良い解か
;			      (progn 
;				(setq smin tau-fgp)
;				(setq k-upper (k-upperbound smin max-fx))))
;			  ;; 深さ優先探索
;			  (get-tau-main f0+g+p f1+g+p f2+g+p g+p rest-plis tau-fgp tau-g+p)))))))
;	    ;; end of definition of local functions
;	    (if (<= n *tau-table-size-base*)
;		(svref *tau-table* f)
;		(multiple-value-bind (f0 f1 f2) (subfuncs f n)
;		  (let* ((term-lis (svref *1term-multi-array* n-1))
;			 (g 0)
;			 (tau-g 0)
;			 (tau-f0 (get-tau f0 n-1))
;			 (tau-f1 (get-tau f1 n-1))
;			 (tau-f2 (get-tau f2 n-1))
;			 (tau-fg (sum-smaller3 tau-f0 tau-f1 tau-f2 tau-g)))
;		    (setq smin tau-fg)
;		    (setq max-fx (max tau-f0 tau-f1 tau-f2))
;		    (setq k-upper (k-upperbound smin max-fx))
;		    (get-tau-main f0 f1 f2 g term-lis tau-fg 0)
;		    smin))))))


;;;
;;; (多出力関数の)積項リストを生成するための関数
;;;

(defun set-1term-multi (n m &key (order nil))
  ;; あらかじめ積項リストを計算しておいて、配列に格納する。
  (setq *1term-multi-array* (make-array (1+ n)))
  (setf (svref *1term-multi-array* 0) (make-1term-multi 0 m))
  (do ((i 1 (1+ i)))
      ((= i (1+ n)) 'ok)
      (let ((result (mapcan #'(lambda (term) (extend-term term i m))
			    (svref *1term-multi-array* (1- i)))))
	(if order ; 積項リストをソートしたい場合
	    (setf (svref *1term-multi-array* i) 
		  (sort-1term-list result :order order))
	    (setf (svref *1term-multi-array* i) 
		  result)))))

(defun extend-term (term n m)
  ;; (n-1)変数積項termから、３つのn変数積項関数を生成してリストで返す。
  ;; termにつけるx_nのリテラルとしては、
  ;; 否定リテラル、肯定リテラル、定数1リテラルの３とおりである。
  ;; (termは関数0ではないとする)
  m ; (reserved)
  (let* ((bit-length-of-subfunc (svref *bit-length-of-func-array* (1- n)))
	 (neg-term term)
	 (pos-term (ash term bit-length-of-subfunc))
	 (con-term (logxor neg-term pos-term)))
    (list con-term pos-term neg-term))) ; 並び順には意味はない

(defun make-1term-multi (n m)
  ;; n変数積項関数のリストを求める。
  ;; (関数0は積項関数でないとする。)
  (if (= n 0)
      (make-m-literal-list m)
      (mapcan #'(lambda (term)
		  (extend-term term n m))
	      (make-1term-multi (1- n) m))))

(defun make-m-literal-list (m)
  ;; (0,m)関数のうち、関数0以外のすべての関数をリストにして返す。(m >= 1)
  (let ((result nil))
    (dotimes (i (expt2 m) (cdr (reverse result)))
      (push i result))))
	
;;; 1termのリストをsortするときに使う

(defun sort-1term-list (1term-list &key (order 'up))
  ;; 1term-listをセル数順にソートしたリストを返す。
  ;; :orderキーに'upまたは'downを指定することで、
  ;; それぞれ昇順または降順になる。
  ;; セル数が等しい積項があるので、ソートの結果は一意ではない。
  (cond ((equal order 'up)
	 (sort 1term-list #'bint-area-leq))
	((equal order 'down)
	 (sort 1term-list #'bint-area-geq))
	(t
	 (error "`:order' should be 'up or 'down "))))

(defun bint-area-leq (bint1 bint2)
  ;; bint1とbint2のセル数（最小積項数）を比較し、
  ;; bint1 <= bint2 かどうか返す。
  (<= (logcount bint1) (logcount bint2)))

(defun bint-area-geq (bint1 bint2)
  ;; bint1とbint2のセル数（最小積項数）を比較し、
  ;; bint1 >= bint2 かどうか返す。
  (>= (logcount bint1) (logcount bint2)))


;;;
;;; 以下は実験用関数。
;;;

(defun print-debug-data (n tau-g len smin tau-fgp)
  ;; デバッグデータの出力
  (format t "~%~a n=~a~a k=~a, rest=~a, tau-fgp=~a~a, lower-bound=~a."
	  (make-space (* (- n 5) 2))
	  n
	  (make-space tau-g)
	  (1+ tau-g)
	  len
	  tau-fgp
	  (if (= smin tau-fgp) "*" "")
	  (- tau-fgp (- (floor (/ (1- smin) 3)) tau-g))))

(defun test-all-functions (n)
  ;; get-tauの動作チェック用手続き
  ;; すべての(n,m)関数について、get-tauを実行し、結果をリストで返す。
  (let ((result nil))
    (dotimes (i (svref *num-of-funcs-array* n) (reverse result))
      (push (list i (get-tau i n)) result))))

(defun make-space (n)
  ;; おまけ関数
  ;; 指定した長さの空白を生成する。
  ;; テストモードで使用している。
  (if (<= n 0)
      ""
      (concatenate 'string " " (make-space (1- n)))))

(defun up-down-compare (f n)
  ;; term-listがupのときとdownのときとで解の計算時間が
  ;; どの程度異なるかを確認する。
  (setq *top-var* n)
  (setq *counter* 0)
  (set-1term-multi n *multi* :order 'up)
  (time (get-tau f n))
  (print (list 'up-count *counter*))
  (setq *counter* 0)
  (set-1term-multi n *multi* :order 'down)
  (time (get-tau f n))
  (print (list 'down-count *counter*))
)


;(defvar max-n 6)
;(defvar multi 1)
;(set-size max-n multi)
;(set-table 4)


